#ifndef VECTOR_H_
#define VECTOR_H_

#include <cmath>
#include <string>

namespace arda 
   {
   namespace Vector 
      {

      ////////////////////////////////////////////////////////////////////////////////
      // Vectors
      //
      // Supported operations on vectors
      //
      // assign*()	Assign the value of one vector to another.
      // add*()		Add two vectors
      // cross()	Cross product of two 3-space vectors.
      // dot*()		Dot product.
      // get_angle*()	Calculate angle between two vectors.
      // length*()	Length of a vector (the Euclidean norm).
      // normalize*()	Normalize the vector.
      // proj*()	Calculate the projection of one vector onto another.
      // scale*()	Multiply vector by a scalar.
      // subtract*()	A convenience function to avoid writing add(v1, scale(v2, -1))
      // getstring*()	Returns a printable string representation of the vector.
      //
      // All functions except getstring are defined inline.

      typedef int Vector2i[2];
      typedef float Vector2f[2];
      typedef double Vector2d[2];

      typedef int Vector3i[3];
      typedef float Vector3f[3];
      typedef double Vector3d[3];

      typedef int Vector4i[4];
      typedef float Vector4f[4];
      typedef double Vector4d[4];

      //////////////////////////////////////////////////////////////////////////
      // getstring*(Vector v)
      //	Returns a C string representation of the vector.
      //	This is one of the few non-inline Vector functions.

      std::string & getstring2v(Vector2i v, std::string & s);
      std::string & getstring2v(Vector2f v, std::string & s);
      std::string & getstring2v(Vector2d v, std::string & s);
      std::string & getstring3v(Vector3i v, std::string & s);
      std::string & getstring3v(Vector3f v, std::string & s);
      std::string & getstring3v(Vector3d v, std::string & s);
      std::string & getstring4v(Vector4i v, std::string & s);
      std::string & getstring4v(Vector4f v, std::string & s);
      std::string & getstring4v(Vector4d v, std::string & s);
      
      //////////////////////////////////////////////////////////////////////////
      // scale*(Vector v, scalar s)
      //	Multiply a vector by a scalar.
      //	Returns v so that the result may be passed to other functions.

      inline int * scale2v(Vector2i v, const int s)
	 { v[0] *= s ; v[1] *= s; return v; }
      inline int * scale2v(Vector2i v, const float s)
	 { v[0] *= s ; v[1] *= s; return v; }
      inline int * scale2v(Vector2i v, const double s)
	 { v[0] *= s ; v[1] *= s; return v; }

      inline float * scale2v(Vector2f v, const int s)
	 { v[0] *= s ; v[1] *= s; return v; }
      inline float * scale2v(Vector2f v, const float s)
	 { v[0] *= s ; v[1] *= s; return v; }
      inline float * scale2v(Vector2f v, const double s)
	 { v[0] *= s ; v[1] *= s; return v; }

      inline double * scale2v(Vector2d v, const int s)
	 { v[0] *= s ; v[1] *= s; return v; }
      inline double * scale2v(Vector2d v, const float s)
	 { v[0] *= s ; v[1] *= s; return v; }
      inline double * scale2v(Vector2d v, const double s)
	 { v[0] *= s ; v[1] *= s; return v; }


      inline int * scale3v(Vector3i v, const int s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
      inline int * scale3v(Vector3i v, const float s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
      inline int * scale3v(Vector3i v, const double s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
 
      inline float * scale3v(Vector3f v, const int s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
      inline float * scale3v(Vector3f v, const float s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
      inline float * scale3v(Vector3f v, const double s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }

      inline double * scale3v(Vector3d v, const int s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
      inline double *  scale3v(Vector3d v, const float s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }
      inline double * scale3v(Vector3d v, const double s)
	 { v[0] *= s ; v[1] *= s; v[2] *= s; return v; }


      inline int * scale4v(Vector4i v, const int s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }
      inline int * scale4v(Vector4i v, const float s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }
      inline int * scale4v(Vector4i v, const double s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }

      inline float * scale4v(Vector4f v, const int s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }
      inline float * scale4v(Vector4f v, const float s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }
      inline float * scale4v(Vector4f v, const double s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }

      inline double * scale4v(Vector4d v, const int s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }
      inline double * scale4v(Vector4d v, const float s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }
      inline double * scale4v(Vector4d v, const double s) 
	 { v[0] *= s ; v[1] *= s;  v[2] *= s;  v[3] *= s; return v; }


      //////////////////////////////////////////////////////////////////////////
      // dot*(Vector v1, Vector v2)
      //	The dot product of two vectors, which must be the same type
      //	Returns a long for int vectors, a double for floating point vectors.

      inline long dot2v(Vector2i v1, Vector2i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1]; }
      inline double dot2v(Vector2f v1, Vector2f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1]; }
      inline double dot2v(Vector2d v1, Vector2d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1]; }

      inline long dot3v(Vector3i v1, Vector3i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1] + (long) v1[2]*v2[2]; }
      inline double dot3v(Vector3f v1, Vector3f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1] + (double) v1[2]*v2[2]; }
      inline double dot3v(Vector3d v1, Vector3d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]; }

      inline long dot4v(Vector4i v1, Vector4i v2)
	 { return (long) v1[0]*v2[0] + (long) v1[1]*v2[1] + (long) v1[2]*v2[2] + (long) v1[3]*v2[3]; }
      inline double dot4v(Vector4f v1, Vector4f v2)
	 { return (double) v1[0]*v2[0] + (double) v1[1]*v2[1] + (double) v1[2]*v2[2] + (double) v1[3]*v2[3]; }
      inline double dot4v(Vector4d v1, Vector4d v2)
	 { return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2] + v1[3]*v2[3]; }


      //////////////////////////////////////////////////////////////////////////
      // length*(Vector v)
      //	Get the length of a vector.
      //	Returns a float for float for float vectors, and a double for 
      //	int and double vectors.

      inline double length2v(Vector2i v)
	 { return sqrt((dot2v(v, v))); }
      inline double length2v(Vector2f v)
	 { return sqrtf((dot2v(v, v))); }
      inline double length2v(Vector2d v)
	 { return sqrt((dot2v(v, v))); }

      inline double length3v(Vector3i v)
	 { return sqrt((dot3v(v, v))); }
      inline double length3v(Vector3f v)
	 { return sqrtf((dot3v(v, v))); }
      inline double length3v(Vector3d v)
	 { return sqrt((dot3v(v, v))); }

      inline double length4v(Vector4i v)
	 { return sqrt((dot4v(v, v))); }
      inline double length4v(Vector4f v)
	 { return sqrtf((dot4v(v, v))); }
      inline double length4v(Vector4d v)
	 { return sqrt((dot4v(v, v))); }


      //////////////////////////////////////////////////////////////////////////
      // add*(Vector v1, Vector v2)
      //	Add v2 to v1.
      //        Returns v1 so that the result may be used in further 
      //	calculations.

      inline int * add2v(Vector2i v1, Vector2i v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; return v1; }
      inline float * add2v(Vector2f v1, Vector2f v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; return v1; }
      inline double * add2v(Vector2d v1, Vector2d v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; return v1; }

      inline int * add3v(Vector3i v1, Vector3i v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; v1[2] += v2[2]; return v1; }
      inline float * add3v(Vector3f v1, Vector3f v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; v1[2] += v2[2];  return v1; }
      inline double * add3v(Vector3d v1, Vector3d v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; v1[2] += v2[2];  return v1; }

      inline int * add4v(Vector4i v1, Vector4i v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; v1[2] += v2[2]; v1[3] += v2[3]; return v1; }
      inline float * add4v(Vector4f v1, Vector4f v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; v1[2] += v2[2]; v1[3] += v2[3]; return v1; }
      inline double * add4v(Vector4d v1, Vector4d v2)
	 { v1[0] += v2[0]; v1[1] += v2[1]; v1[2] += v2[2]; v1[3] += v2[3]; return v1; }

      //////////////////////////////////////////////////////////////////////////
      // normalize*(Vector v)
      //	Normalizes the vecor.  Only defined for floating point vectors.
      //	Returns the vector.

      inline float * normalize2v(Vector2f v)
	 { double l = length2v(v); v[0] /= l; v[1] /= l; return v; }
      inline double * normalize2v(Vector2d v)
	 { double l = length2v(v); v[0] /= l; v[1] /= l; return v; }

      inline float * normalize3v(Vector3f v)
	 { double l = length3v(v); v[0] /= l; v[1] /= l; v[2] /= l; return v; }
      inline double * normalize3v(Vector3d v)
	 { double l = length3v(v); v[0] /= l; v[1] /= l; v[2] /= l; return v; }

      inline float * normalize4v(Vector4f v)
	 { double l = length4v(v); v[0] /= l; v[1] /= l; v[2] /= l; v[3] /= l; return v; }
      inline double * normalize4v(Vector4d v)
	 { double l = length4v(v); v[0] /= l; v[1] /= l; v[2] /= l; v[3] /= l; return v; }

      //////////////////////////////////////////////////////////////////////////
      // cross(Vector3 v)
      //	Only defined for floating point 3 vectors.
      //	Returns the result vector.

      inline float * cross(Vector3f v1, Vector3f v2, Vector3f vres)
	 { 
	 vres[0] =  v1[1]*v2[2] - v2[1]*v1[2];
	 vres[1] = -v1[0]*v2[2] + v2[0]*v1[2];
	 vres[2] =  v1[0]*v2[1] - v2[0]*v1[1];
	 return vres;
	 }

      inline double * cross(Vector3d v1, Vector3d v2, Vector3d vres)
	 { 
	 vres[0] =  v1[1]*v2[2] - v2[1]*v1[2];
	 vres[1] = -v1[0]*v2[2] + v2[0]*v1[2];
	 vres[2] =  v1[0]*v2[1] - v2[0]*v1[1];
	 return vres;
	 }
   
      //////////////////////////////////////////////////////////////////////////
      // get_angle*(Vector3 v1, Vector3 v2)
      //	Returns the angle in radians between the two vectors.
      //
      // get_angle[234]n() take pre-normalized vectors as arguments and will
      //	fail otherwise.  This version is not defined for interger
      //	vectors since you can't normalize an integer vector.
      // get_angle[234]() take arbitrary vectors as arguments.
      //
      // I started out with two version of get_angle3:
      // get_angle3_1 implemented as
      //  return acos(dot3v(v1, v2)/(length3v(v1)*length3v(v2)));
      // get_angle3_2 implemented as
      //  double tmp = dot3v(v1, v2); 
      //  return acos(sqrtf(tmp*tmp/(dot3v(v1,v1)*dot3v(v2,v2))));
      // 
      // I ran timing tests, since I expected the second one - the one I chose to 
      // use, in the end - to be faster since it only resulted in one sqrt call
      // instead of two.  Here are the results
      //
      // get_angle3_1 44.753006890 seconds for 500,000,000 iterations.
      // get_angle3_2 36.940170102 seconds for 500,000,000 iterations.
      //
      // Faster?  Yes. About 7 nanoseconds vs 9.  A ridiculous optimization
      // to worry about?  Eh.  I might as well use it.

      inline double get_angle2n(Vector2f v1, Vector2f v2)
	 { return acos(dot2v(v1, v2)); }
      inline double get_angle2n(Vector2d v1, Vector2d v2)
	 { return acos(dot2v(v1, v2)); }
      inline double get_angle2v(Vector2i v1, Vector2i v2)
	 { 
	 double tmp = dot2v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot2v(v1,v1)*dot2v(v2,v2)))); 
	 }
      inline double get_angle2v(Vector2f v1, Vector2f v2)
	 { 
	 double tmp = dot2v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot2v(v1,v1)*dot2v(v2,v2)))); 
	 }
      inline double get_angle2v(Vector2d v1, Vector2d v2)
	 { 
	 double tmp = dot2v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot2v(v1,v1)*dot2v(v2,v2)))); 
	 }

      inline double get_angle3n(Vector3f v1, Vector3f v2)
	 { return acos(dot3v(v1, v2)); }
      inline double get_angle3n(Vector3d v1, Vector3d v2)
	 { return acos(dot3v(v1, v2)); }
      inline double get_angle3v(Vector3i v1, Vector3i v2)
	 { 
	 double tmp = dot3v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot3v(v1,v1)*dot3v(v2,v2)))); 
	 }
      inline double get_angle3v(Vector3f v1, Vector3f v2)
	 { 
	 double tmp = dot3v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot3v(v1,v1)*dot3v(v2,v2)))); 
	 }
      inline double get_angle3v(Vector3d v1, Vector3d v2)
	 { 
	 double tmp = dot3v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot3v(v1,v1)*dot3v(v2,v2)))); 
	 }

      inline double get_angle4n(Vector4f v1, Vector4f v2)
	 { return acos(dot4v(v1, v2)); }
      inline double get_angle4n(Vector4d v1, Vector4d v2)
	 { return acos(dot4v(v1, v2)); }
      inline double get_angle4v(Vector4i v1, Vector4i v2)
	 { 
	 double tmp = dot4v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot4v(v1,v1)*dot4v(v2,v2)))); 
	 }
      inline double get_angle4v(Vector4f v1, Vector4f v2)
	 { 
	 double tmp = dot4v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot4v(v1,v1)*dot4v(v2,v2)))); 
	 }
      inline double get_angle4v(Vector4d v1, Vector4d v2)
	 { 
	 double tmp = dot4v(v1, v2); 
	 return acos(sqrtf(tmp*tmp/(dot4v(v1,v1)*dot4v(v2,v2)))); 
	 }

      //////////////////////////////////////////////////////////////////////////
      // subtract*(Vector v1, Vector v2)
      //	Subtract v2 from v1,
      //	Equivalent to add(v1, scale(v2, -1)).
      //        Returns v1 so that the result may be used in further 
      //	calculations.

      inline int * subtract2v(Vector2i v1, Vector2i v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; return v1; }
      inline float * subtract2v(Vector2f v1, Vector2f v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; return v1; }
      inline double * subtract2v(Vector2d v1, Vector2d v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; return v1; }

      inline int * subtract3v(Vector3i v1, Vector3i v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; v1[2] -= v2[2]; return v1; }
      inline float * subtract3v(Vector3f v1, Vector3f v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; v1[2] -= v2[2];  return v1; }
      inline double * subtract3v(Vector3d v1, Vector3d v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; v1[2] -= v2[2];  return v1; }

      inline int * subtract4v(Vector4i v1, Vector4i v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; v1[2] -= v2[2]; v1[3] -= v2[3]; return v1;}
      inline float * subtract4v(Vector4f v1, Vector4f v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; v1[2] -= v2[2]; v1[3] -= v2[3]; return v1; }
      inline double * subtract4v(Vector4d v1, Vector4d v2)
	 { v1[0] -= v2[0]; v1[1] -= v2[1]; v1[2] -= v2[2]; v1[3] -= v2[3]; return v1; }


      //////////////////////////////////////////////////////////////////////////
      // assign*(Vector v1, Vector v2)
      //	Copies the value of v2 into v1.
      //	Returns v1.

      inline int * assign2v(Vector2i v1, Vector2i v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; return v1; }
      inline float * assign2v(Vector2f v1, Vector2f v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; return v1; }
      inline double * assign2v(Vector2d v1, Vector2d v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; return v1; }

      inline int * assign3v(Vector3i v1, Vector3i v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; return v1; }
      inline float * assign3v(Vector3f v1, Vector3f v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; return v1; }
      inline double * assign3v(Vector3d v1, Vector3d v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; return v1; }

      inline int * assign4v(Vector4i v1, Vector4i v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; v1[3] = v2[3]; return v1; }
      inline float * assign4v(Vector4f v1, Vector4f v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; v1[3] = v2[3]; return v1; }
      inline double * assign4v(Vector4d v1, Vector4d v2)
	 { v1[0] = v2[0]; v1[1] = v2[1]; v1[2] = v2[2]; v1[3] = v2[3]; return v1; }


      //////////////////////////////////////////////////////////////////////////
      // proj*(Vector v1, Vector v2, Vector vres)
      //	Calculates the projection of v1 onto v2 and returns the result 
      //	in vres.
      //	Only defined for floating point vectors.
      //	Returns vres.

      inline float * proj2v(Vector2f v1, Vector2f v2, Vector2f vres)
	 { assign2v(vres, v2); scale2v(vres, (dot2v(v1, vres)/dot2v(vres,vres))); return vres; }
      inline double * proj2v(Vector2d v1, Vector2d v2, Vector2d vres)
	 { assign2v(vres, v2); scale2v(vres, (dot2v(v1, vres)/dot2v(vres,vres))); return vres; }

      inline float * proj3v(Vector3f v1, Vector3f v2, Vector3f vres)
	 { assign3v(vres, v2); scale3v(vres, (dot3v(v1, vres)/dot3v(vres,vres))); return vres; }
      inline double * proj3v(Vector3d v1, Vector3d v2, Vector3d vres)
	 { assign3v(vres, v2); scale3v(vres, (dot3v(v1, vres)/dot3v(vres,vres))); return vres; }

      inline float * proj4v(Vector4f v1, Vector4f v2, Vector4f vres)
	 { assign4v(vres, v2); scale4v(vres, (dot4v(v1, vres)/dot4v(vres,vres))); return vres; }
      inline double * proj4v(Vector4d v1, Vector4d v2, Vector4d vres)
	 { assign4v(vres, v2); scale4v(vres, (dot4v(v1, vres)/dot4v(vres,vres))); return vres; }

      } // Vector

   } // namespace arda

#endif // VECTOR_H_
